# Copyright the author(s) of intc.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
import sys
from typing import Dict, List, Optional, Union

from tree_sitter import Language, Node, Parser

if sys.platform == "win32":
    sys_post_fix = "win"
elif sys.platform == "darwin":
    sys_post_fix = "mac"
else:
    sys_post_fix = "linux"


class YamlParser(object):
    """docstring for ParserYAML"""

    def __init__(self):
        super(YamlParser, self).__init__()
        self._parser = Parser()
        dynamic_lib_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
            "lib",
            f"yaml_{sys_post_fix}_ts.so",
        )
        self._parser.set_language(Language(dynamic_lib_path, "yaml"))
        self.skip = {"comment"}

    def parser(self, doc: str) -> Optional[Union[Dict, List]]:
        """parser the json string to AST
        Args:
            doc: the source document

        Returns:
            the parser tree
        """
        return self.parser_object(self._parser.parse(bytes(doc, "utf8")).root_node)

    @staticmethod
    def print_parser_tree(node: Node, deep=0) -> None:
        for children in node.named_children:
            print(deep * "    " + children.type)
            YamlParser.print_parser_tree(children, deep + 1)

    def drop_quote(self, _text: bytes) -> str:
        text = _text.decode()
        if (
            len(text) >= 2
            and (text[0] == '"' and text[-1] == '"')
            or (text[0] == "'" and text[-1] == "'")
        ):
            return text[1:-1]
        return text

    def parser_pair(self, node: Node, deep: int) -> Dict:
        key = self.parser_object(node.named_children[0])
        key = self.parser_object(node.named_children[0], deep + 1)
        if len(node.named_children) == 2:
            value = self.parser_object(node.named_children[1], deep + 1)
        else:
            assert len(node.named_children) == 1
            value = None
        result = {}
        result["__type"] = "pair"
        result["__range"] = (node.start_point, node.end_point)
        result["__value"] = value
        result["__key"] = key
        return result

    def parser_flow_node(self, node: Node, deep: int):
        assert len(node.named_children) == 1, node.named_children
        return self.parser_object(node.named_children[0], deep + 1)

    def parser_plain_scalar(self, node: Node, deep: int):
        assert len(node.named_children) == 1, node.named_children
        return self.parser_object(node.named_children[0], deep + 1)

    def parser_string(self, node: Node, deep: int) -> Dict:
        assert len(node.named_children) == 0, node.named_children
        return {
            "__type": "string",
            "__value": self.drop_quote(node.text),
            "__range": (node.start_point, node.end_point),
        }

    def parser_number(self, node: Node, deep: int) -> Dict:
        assert len(node.named_children) == 0, node.named_children
        return {
            "__type": "number",
            "__value": eval(node.text.decode()),
            "__range": (node.start_point, node.end_point),
        }

    def parser_array(self, node: Node, deep: int) -> Dict:
        return {
            "__type": "array",
            "__value": [self.parser_object(child) for child in node.named_children],
            "__range": (node.start_point, node.end_point),
        }

    def parser_block_node(self, node: Node, deep: int):
        assert len(node.named_children) == 1, node.named_children
        return self.parser_object(node.named_children[0], deep + 1)

    def parser_null_true_false(self, node: Node, deep: int):
        type_map = {
            "null_scalar": "null",
            "boolean_scalar": "bool",
        }
        return {
            "__type": type_map[node.type],
            "__value": "null_or_bool",
            "__range": (node.start_point, node.end_point),
        }

    def parser_object(self, node: Node, deep: int = 0):
        if deep == 0 and node.type == "ERROR":
            parser_childs = [
                self.parser_object(child)
                for child in node.named_children
                if child.type not in self.skip
            ]
            return parser_childs
        if node.type in self.skip:
            return None
        if node.type == "block_mapping_pair":
            return self.parser_pair(node, deep + 1)

        if node.type == "flow_sequence":
            return self.parser_array(node, deep + 1)
        if node.type == "plain_scalar":
            return self.parser_plain_scalar(node, deep + 1)
        if node.type == "flow_node":
            return self.parser_flow_node(node, deep + 1)
        if node.type in {"string_scalar", "double_quote_scalar", "single_quote_scalar"}:
            return self.parser_string(node, deep + 1)

        if node.type in {"integer_scalar", "float_scalar"}:
            return self.parser_number(node, deep + 1)

        if node.type in {"boolean_scalar", "null_scalar"}:
            return self.parser_null_true_false(node, deep + 1)

        if node.type == "stream":
            for child in node.named_children:
                result = self.parser_object(child, deep + 1)
                if result is not None:
                    return result
            return {}

        if node.type == "block_node":
            return self.parser_block_node(node, deep + 1)
        if node.type == "block_mapping":
            values = []
            for child in node.named_children:
                if child.type in self.skip:
                    continue
                value = self.parser_object(child, deep + 1)
                if value:
                    values.append(value)
            return values

        if node.type == "block_sequence":
            return {
                "__type": "array",
                "__value": [
                    self.parser_object(child, deep) for child in node.named_children
                ],
                "__range": (node.start_point, node.end_point),
            }
        if node.type == "block_sequence_item":
            return self.parser_object(node.named_children[0], deep + 1)

        if node.type == "document":
            result = []
            for child in node.named_children:
                if child.type in self.skip:
                    continue
                result.append(
                    {
                        "__type": "document",
                        "__value": self.parser_object(child, deep + 1),
                        "__range": (child.start_point, child.end_point),
                    }
                )
            return result
        return {}


if __name__ == "__main__":
    parser = YamlParser()
    print(
        parser.parser(
            """
    # type=dlk

    "@config":
        "_name_": basic
        "@optimizer@adam": 1
        name:
            - a
            - b
    """
        )
    )



# Copyright the author(s) of intc.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

"""Intc Language Server command line interface."""

import argparse
import logging
import os
import sys
from logging.handlers import RotatingFileHandler
from textwrap import dedent

from platformdirs import user_log_dir

from intc_lsp.server import intc_server
from intc_lsp.version import __version__

log_dir = user_log_dir("intc_lsp")


def get_version() -> str:
    """Get the program version."""
    return __version__


def cli() -> None:
    """intc language server cli entrypoint."""
    parser = argparse.ArgumentParser(
        prog="intc-language-server",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="intc language server: an LSP wrapper for intc.",
        epilog=dedent(
            """\
            Examples:

                Run over stdio     : intc-language-server
                Run over tcp       : intc-language-server --tcp
                Run over websockets:
                    # only need to pip install once per env
                    pip install pygls[ws]
                    intc-language-server --ws

            Notes:

                For use with web sockets, user must first run
                'pip install pygls[ws]' to install the correct
                version of the websockets library.
            """
        ),
    )
    parser.add_argument(
        "--version",
        help="display version information and exit",
        action="store_true",
    )
    parser.add_argument(
        "--tcp",
        help="use TCP web server instead of stdio",
        action="store_true",
    )
    parser.add_argument(
        "--ws",
        help="use web socket server instead of stdio",
        action="store_true",
    )
    parser.add_argument(
        "--host",
        help="host for web server (default 127.0.0.1)",
        type=str,
        default="127.0.0.1",
    )
    parser.add_argument(
        "--port",
        help="port for web server (default 9999)",
        type=int,
        default=9999,
    )
    parser.add_argument(
        "--log-file",
        default=os.path.join(log_dir, "log"),
        help="redirect logs to file specified",
        type=str,
    )
    parser.add_argument(
        "--log_level",
        default=2,
    )
    args = parser.parse_args()
    if args.version:
        print(get_version())
        sys.exit(0)
    if args.tcp and args.ws:
        print(
            "Error: --tcp and --ws cannot both be specified",
            file=sys.stderr,
        )
        sys.exit(1)
    log_level = {
        0: logging.ERROR,
        1: logging.WARN,
        2: logging.INFO,
        3: logging.DEBUG,
    }.get(
        args.log_level,
        logging.WARN,
    )
    logger = logging.getLogger()
    logger_intc = logging.getLogger("intc_lsp")

    if args.log_file:
        if not os.path.isfile(os.path.dirname(args.log_file)):
            os.makedirs(os.path.dirname(args.log_file), exist_ok=True)
            logger.setLevel(log_level)
            logger_intc.setLevel(log_level)
            file_handler = RotatingFileHandler(
                filename=args.log_file,
                mode="w",
                encoding="utf8",
                maxBytes=5 * 1024 * 1024,
            )
            file_formatter = logging.Formatter(
                fmt="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
                datefmt="%m/%d/%Y %H:%M:%S",
            )
            file_handler.setFormatter(file_formatter)
            logger.addHandler(file_handler)
    else:
        logging.basicConfig(stream=sys.stderr, level=log_level)

    if args.tcp:
        intc_server.start_tcp(host=args.host, port=args.port)
    elif args.ws:
        intc_server.start_ws(host=args.host, port=args.port)
    else:
        intc_server.start_io()


if __name__ == "__main__":
    cli()



# Copyright the author(s) of intc.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import importlib
import io
import json
import logging
import os
import re
import sys
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple
from urllib.parse import urlparse

import hjson
from intc import ic_repo
from intc.loader import load_submodule
from lsprotocol.types import (
    ALL_TYPES_MAP,
    INITIALIZE,
    INITIALIZED,
    TEXT_DOCUMENT_CODE_ACTION,
    TEXT_DOCUMENT_COMPLETION,
    TEXT_DOCUMENT_DEFINITION,
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_SAVE,
    TEXT_DOCUMENT_FORMATTING,
    TEXT_DOCUMENT_HOVER,
    CodeAction,
    CodeActionKind,
    CodeActionOptions,
    CodeActionParams,
    Command,
    CompletionItem,
    CompletionItemKind,
    CompletionList,
    CompletionOptions,
    CompletionParams,
    CompletionTriggerKind,
    Diagnostic,
    DiagnosticSeverity,
    DocumentFormattingParams,
    Hover,
    InitializeParams,
    InitializeResult,
    InitializeResultServerInfoType,
    Location,
    MarkupContent,
    MarkupKind,
    Position,
    Range,
    SaveOptions,
    ServerCapabilities,
    TextDocumentPositionParams,
    TextDocumentSaveRegistrationOptions,
    TextDocumentSyncKind,
    TextEdit,
    WorkspaceEdit,
)
from pygls.protocol import lsp_method
from pygls.server import LanguageServer

from intc_lsp.src import HoverType, IntcResolve
from intc_lsp.version import __version__

logger = logging.getLogger("intc_lsp")


ALL_TYPES_MAP["TextDocumentSaveOptions"] = TextDocumentSaveRegistrationOptions


class IntcLanguageServer(LanguageServer):
    """The intc LanguageServer"""

    def __init__(
        self,
        name,
        version,
        text_document_sync_kind=TextDocumentSyncKind.Full,
        max_workers=4,
    ) -> None:
        super().__init__(
            name=name,
            version=version,
            text_document_sync_kind=text_document_sync_kind,
            max_workers=max_workers,
        )
        self.did_opend_files = set()
        self.resolve: IntcResolve = None
        self.options = {}
        self.modules_pattern = ""
        self.support_file_types = ["json", "yaml", "yml", "jsonc", "hjson", "json5"]
        self.entry_pattern = ""

    def update_config_partern(self, root_path):
        """update the pattern of the entry and modules
        Args:
            root_path:
                the root path of the workspace
        Returns:
            None
        """
        if not self.options:
            return
        entry_pattern = ""
        modules_pattern = ""
        for entry in self.options.get("entry", []):
            entry_path = os.path.join(
                root_path, entry, r"[^/]*\." + f"({'|'.join(self.support_file_types)})"
            )
            entry_pattern = (
                f"{entry_path}"
                if not entry_pattern
                else f"{entry_pattern}|{entry_path}"
            )
        for module in self.options.get("module", []):
            module_path = os.path.join(
                root_path, module, r"[^/]*\." + f"({'|'.join(self.support_file_types)})"
            )
            modules_pattern = (
                f"{module_path}"
                if not modules_pattern
                else f"{modules_pattern}|{module_path}"
            )
        logger.info(
            f"update_config_partern: entry_pattern : {entry_pattern}, modules_pattern : {modules_pattern}"
        )
        self.entry_pattern = re.compile(entry_pattern)
        self.modules_pattern = re.compile(modules_pattern)

    def init_new_file(self, params):
        uri = urlparse(params.text_document.uri).path
        if uri in self.did_opend_files:
            return
        logger.info(f"init_new_file: {uri}")
        try:
            root_path = Path(self.workspace.root_path)
        except Exception as e:
            root_path = Path(uri).parent

        intc_setting = ""
        real_root = Path(uri).parent
        while (root_path.parent != real_root) and not intc_setting:
            for meta_config in [".intc.json", ".intc.jsonc"]:
                if os.path.isfile(os.path.join(real_root, meta_config)):
                    intc_setting = os.path.join(real_root, meta_config)
                    break
            if not intc_setting:
                real_root = real_root.parent

        if not intc_setting:
            logger.info(f"self: can not found init setting for {uri}")
            return
        try:
            self.options = dict(hjson.load(open(intc_setting, "r")))
            self.update_config_partern(real_root)
            logger.info(f"init: load .intc.json done")
            logger.info(json.dumps(self.options, indent=4))
        except Exception as e:
            logger.error(f"init: load .intc.json error : {e}")
            self.show_message_log(f"intc_server: load .intc.json error")

        try:
            self.resolve = IntcResolve(intc_server)
            logger.info(f"init: init the intc_resolve")
        except Exception as e:
            logger.error(f"init: init intc_resolve error : {e}")
            self.show_message_log(f"intc_server: init intc_resolve error")

        logger.info(f"init: before dynamic import")
        logger.info(f"init: ic_repo: {ic_repo}")

        null_buffer = io.StringIO()
        original_stdout = sys.stdout
        original_stderr = sys.stderr
        sys.stdout = null_buffer
        sys.stderr = null_buffer

        try:
            sys.path.append(str(real_root))
        except:
            logger.error(f"init: add root path {real_root} error")
        os.environ["IN_INTC"] = "1"
        for package in self.options.get("src", []):
            try:
                importlib.import_module(package)
                logger.info(f"init: import {package}")
            except Exception as e:
                logger.error(f"init: on import package `{package}` error : {e}")

        try:
            logger.info(load_submodule(str(real_root)))
        except Exception as e:
            logger.info(f"loader sub module error: {e}")

        sys.stdout = original_stdout
        sys.stderr = original_stderr
        null_buffer.seek(0)
        output = null_buffer.read()
        logger.info(f"init import output: {output}")
        logger.info(f"init: after dynamic import")
        logger.info(f"init: ic_repo: {ic_repo}")
        logger.info(f"init: init the intc_lsp done")
        self.did_opend_files.add(uri)


intc_server = IntcLanguageServer("intc-language-server", __version__)


@intc_server.feature(INITIALIZED)
def initialize(params: InitializeParams) -> Optional[InitializeResult]:
    """Initialize the language server
    Args:
        params:
            the init parameters provide by the client
    Returns:
        Optional[InitializeResult]
    """
    logger.info(f"init: paras: {params}")
    intc_server.show_message_log(f"intc_server: start init")
    if not intc_server.workspace.root_path:
        intc_server.show_message_log(f"init: no root path")
        return
    else:
        logger.info(f"init: root path: {intc_server.workspace.root_path}")


@intc_server.feature(
    TEXT_DOCUMENT_COMPLETION,
    CompletionOptions(trigger_characters=['"', ":", " ", "@"]),
)
def completions(params: CompletionParams) -> CompletionList:
    """Provide completion list trigger by the trigger_characters
    Args:
        params:
            the parameters provide by the client, provide the completion position and the source uri
    Returns:
        a list of completion items
    """
    logger.info(f"completions: paras: {params}")
    completions = intc_server.resolve.completions(
        params.position, params.text_document.uri, params.context.trigger_character
    )
    return completions


@intc_server.feature(TEXT_DOCUMENT_HOVER)
def hover(params: TextDocumentPositionParams) -> Optional[Hover]:
    """Provide the code help information of the hover position
    Args:
        server:
            the intc LanguageServer
        params:
            the parameters provide by the client, provide the hover position and the source uri
    Returns:
        a list of completion items
    """
    intc_server.show_message_log(f"on hover: {params}")

    logger.info(f"hover: paras: {params}")
    hover_result = intc_server.resolve.hover(params.position, params.text_document.uri)
    if hover_result["type"] not in {
        HoverType.RESOLVE_ERROR,
        HoverType.UN_COVER_ERROR,
        HoverType.CURSOR_WORD_NOT_FOUND,
    }:
        return Hover(
            contents=MarkupContent(
                kind=MarkupKind.Markdown, value=hover_result["message"]
            ),
            range=hover_result["range"],
        )
    else:
        return None


@intc_server.feature(TEXT_DOCUMENT_DEFINITION)
def definition(params: TextDocumentPositionParams) -> Optional[List[Location]]:
    """Provide completion list
    Args:
        server:
            the intc LanguageServer
        params:
            the parameters provide by the client, provide the hover position and the source uri
    Returns:
        a list of completion items
    """

    logger.info(f"definition: paras: {params}")
    definitions = intc_server.resolve.definition(
        params.position, params.text_document.uri
    )
    if definitions:
        return definitions


def diagnostics(params: TextDocumentPositionParams) -> None:
    """publish the diagnostics hint
    Args:
        server:
            the intc LanguageServer
        params:
            the parameters provide by the client, provide the source uri
    Returns:
        None
    """

    logger.info(f"diagnostics: paras: {params}")
    try:
        display_diagnostics = intc_server.resolve.diagnostics(params.text_document.uri)
    except Exception as e:
        logger.error(f"diagnostics: error : {e}")
        return
    logger.info(f"display: {display_diagnostics}")
    intc_server.publish_diagnostics(params.text_document.uri, display_diagnostics)


@intc_server.feature(TEXT_DOCUMENT_DID_OPEN)
def did_open(params):
    logger.info(f"did_open: paras: {params}")
    intc_server.init_new_file(params)
    diagnostics(params)


@intc_server.feature(TEXT_DOCUMENT_DID_CHANGE)
def did_change(params):
    logger.info(f"did_change: paras: {params}")
    diagnostics(params)


if __name__ == "__main__":
    intc_server.start_io()



# Copyright the author(s) of intc.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from intc_lsp.src.resolve import HoverType, IntcResolve

__all__ = ["HoverType", "IntcResolve"]



# Copyright the author(s) of intc.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging

logger = logging.getLogger("intc_lsp")


def position_is_in_range(position_range: tuple, position: tuple) -> bool:
    """Check if a given position is within a specified range.

    Args:
        position_range: A tuple representing the start and end positions of a range.
        position: A tuple representing the position (row, col)

    Returns:
        bool: True if the position is within the range, False otherwise.
    """
    range_start = position_range[0][0] * 1e8 + position_range[0][1]
    range_end = position_range[1][0] * 1e8 + position_range[1][1]
    position = position[0] * 1e8 + position[1]
    return position >= range_start and position < range_end


def root_trace(result, position, trace="") -> tuple:
    """Trace the root of a tree structure based on a given position.

    Args:
        result: The result of the tracing process.
        position: The position to trace.

    Returns:
        tuple: A tuple containing the semantic trace, lexical trace, trace value, a flag indicating if the position matches the key or value, anchor dictionary, and additional information.
    """
    anchor_dict = {}
    additional = {}
    if not result:
        return None, None, None, None, anchor_dict, additional
    for node in result:
        semantic_trace, lex_trace, trace_result, is_key_or_none = node_trace(
            node, position, anchor_dict, additional, trace
        )
        if is_key_or_none is not None:
            return (
                semantic_trace,
                lex_trace,
                trace_result,
                is_key_or_none,
                anchor_dict,
                additional,
            )
    return None, None, None, None, anchor_dict, additional


def node_trace_document(
    node, position, anchor_dict: dict, additional: dict, trace: str = ""
):
    """Trace a document node in a tree structure based on a given position.

    Args:
        node: The document node to trace.
        position: The position to trace.
        anchor_dict: A dictionary to store anchor names and their corresponding traces.
        additional: Additional information to be passed along during the tracing process.
        trace: The current trace string. Defaults to an empty string.

    Returns:
        tuple: A tuple containing the semantic trace, lexical trace, trace value, and a flag indicating if the position matches the key or value.
    """
    if not position_is_in_range(node["__range"], position):
        return None, None, None, None
    if trace:
        node["__type"] = "pair"
        node["__key"] = {
            "__type": "string",
            "__value": f"@{trace.lstrip('@')}",
            "__range": ((-1, -1), (-1, -1)),
        }
        return node_trace_pair(node, position, anchor_dict, additional, trace)
    for value in node["__value"]:
        value_result = node_trace(value, position, anchor_dict, additional, trace)
        if value_result[3] is not None:
            return value_result
    return None, None, None, None


def node_trace(
    node, position, anchor_dict: dict, additional: dict, trace: str = ""
) -> tuple:
    """Trace a node in a tree structure based on a given position.

    Args:
        node: The node to trace.
        position: The position to trace.
        anchor_dict: A dictionary to store anchor names and their corresponding traces.
        additional: Additional information to be passed along during the tracing process.
        trace: The current trace string. Defaults to an empty string.

    Returns:
        tuple: A tuple containing the semantic trace, lexical trace, trace value, and a flag indicating if the position matches the key or value.
    """
    if node.get("__type", None) == "pair":
        return node_trace_pair(node, position, anchor_dict, additional, trace)
    elif node.get("__type", None) == "document":
        return node_trace_document(node, position, anchor_dict, additional, trace)
    elif node.get("__type", None) == "array":
        return node_trace_array(node, position, anchor_dict, additional, trace)
    else:
        return node_trace_other(node, position, anchor_dict, additional, trace)


def node_trace_other(
    node, position, anchor_dict: dict, additional: dict, trace: str = ""
) -> tuple:
    """Trace a node of type other in a tree structure based on a given position.

    Args:
        node: The node to trace.
        position: The position to trace.
        anchor_dict: A dictionary to store anchor names and their corresponding traces.
        additional: Additional information to be passed along during the tracing process.
        trace: The current trace string. Defaults to an empty string.

    Returns:
        tuple: A tuple containing the semantic trace, lexical trace, trace value, and a flag indicating if the position matches the key or value.
    """
    if not position_is_in_range(node["__range"], position):
        return None, None, None, None
    return "", "", node.get("__value", None), False


def node_trace_array(
    node, position, anchor_dict: dict, additional: dict, trace: str = ""
):
    """Trace an array node in a tree structure based on a given position.

    Args:
        node: The array node to trace.
        position: The position to trace.
        anchor_dict: A dictionary to store anchor names and their corresponding traces.
        additional: Additional information to be passed along during the tracing process.
        trace: The current trace string. Defaults to an empty string.

    Returns:
        tuple: A tuple containing the semantic trace, lexical trace, trace value, and a flag indicating if the position matches the key or value.
    """
    if not position_is_in_range(node["__range"], position):
        return None, None, None, None
    for i, child in enumerate(node["__value"]):
        if not position_is_in_range(child["__range"], position):
            continue
        semantic_trace, lex_trace, trace_value, is_key_or_none = node_trace(
            child, position, anchor_dict, additional, trace + f".{i}"
        )
        if semantic_trace:
            return (
                f"{i}.{semantic_trace}",
                f"{i}.{lex_trace}",
                trace_value,
                is_key_or_none,
            )
    return None, None, None, None


def node_trace_pair(
    node, position, anchor_dict: dict, additional: dict, trace: str = ""
):
    """Trace a pair node in a tree structure based on a given position.

    Args:
        node: The pair node to trace.
        position: The position to trace.
        anchor_dict: A dictionary to store anchor names and their corresponding traces.
        additional: Additional information to be passed along during the tracing process.
        trace: The current trace string. Defaults to an empty string.

    Returns:
        tuple: A tuple containing the semantic trace, lexical trace, trace value, and a flag indicating if the position matches the key or value.
    """
    trace_value = None
    is_key_or_none = (
        None  # True or False means matching the key or value, None means not match
    )
    on_module_root = False
    semantic_trace = ""
    lex_trace = ""
    if not position_is_in_range(node["__range"], position):
        return None, None, None, None
    module_name = node["__key"]["__value"]
    semantic_module_name = module_name
    is_module = False
    if node.get("__key", {}).get("__value", "").startswith("@"):
        is_module = True

    if position_is_in_range(node["__key"]["__range"], position):
        if not is_module:
            return "", "", node.get("__key", {}).get("__value", ""), True
        else:
            on_module_root = True
            additional["on_semantic_module_name"] = module_name
            additional["on_module_name"] = module_name
            if (not isinstance(node.get("__value", None), list)) or (
                not node.get("__value", None)
            ):
                return "", "", node.get("__key", {}).get("__value", ""), True

    sub_modules = node.get("__value", [])
    if isinstance(sub_modules, dict):
        semantic_trace, lex_trace, trace_value, is_key_or_none = node_trace(
            sub_modules, position, anchor_dict, additional, trace + "." + module_name
        )
        if not on_module_root:
            if is_key_or_none is not None:
                return (
                    semantic_module_name
                    if not semantic_trace
                    else f"{semantic_module_name}.{semantic_trace}",
                    module_name if not lex_trace else f"{module_name}.{lex_trace}",
                    trace_value,
                    False,
                )
            return None, None, None, None
    assert isinstance(sub_modules, list), sub_modules
    for sub_module in sub_modules:
        if (
            is_module
            and sub_module.get("__type", None) == "pair"
            and (sub_module.get("__key", {}).get("__value", "") in {"_name", "_base"})
        ):
            base_name = sub_module.get("__value", {}).get("__value", "")
            if base_name and isinstance(base_name, str):
                semantic_module_name = f"{module_name.lstrip('@')}@{base_name}"
                if semantic_module_name:
                    semantic_module_name = f"@{semantic_module_name}"
                if on_module_root:
                    additional["on_semantic_module_name"] = semantic_module_name
                    additional["on_module_name"] = module_name

        elif (
            is_module
            and sub_module.get("__type", None) == "pair"
            and sub_module.get("__key", {}).get("__value", "") == "_anchor"
        ):
            anchor_name = sub_module.get("__value", {}).get("__value", "")
            if anchor_name and isinstance(anchor_name, str):
                anchor_dict[anchor_name] = (
                    trace + "." + module_name if trace else module_name
                )
        if position_is_in_range(sub_module["__range"], position):
            semantic_trace, lex_trace, trace_value, is_key_or_none = node_trace(
                sub_module, position, anchor_dict, additional, trace + "." + module_name
            )
    if not on_module_root:
        if is_key_or_none is not None:
            semantic_trace = (
                f"{semantic_module_name}.{semantic_trace}"
                if semantic_trace
                else semantic_module_name
            )
            lex_trace = f"{module_name}.{lex_trace}" if lex_trace else module_name
            return semantic_trace, lex_trace, trace_value, is_key_or_none
    else:
        return "", "", node.get("__key", {}).get("__value", ""), True
    return None, None, None, None


if __name__ == "__main__":
    result = [
        {
            "__type": "pair",
            "__value": [
                {
                    "__type": "pair",
                    "__range": ((1, 4), (1, 24)),
                    "__value": {
                        "__type": "string",
                        "__value": "config_a",
                        "__range": ((1, 14), (1, 24)),
                    },
                    "__key": {
                        "__type": "string",
                        "__value": "_base",
                        "__range": ((1, 4), (1, 12)),
                    },
                },
                {
                    "__type": "pair",
                    "__range": ((2, 4), (10, 5)),
                    "__value": [
                        {
                            "__type": "pair",
                            "__range": ((3, 8), (3, 27)),
                            "__value": {
                                "__type": "string",
                                "__value": "child_a",
                                "__range": ((3, 18), (3, 27)),
                            },
                            "__key": {
                                "__type": "string",
                                "__value": "_base",
                                "__range": ((3, 8), (3, 16)),
                            },
                        },
                        {
                            "__type": "pair",
                            "__range": ((4, 8), (4, 31)),
                            "__value": {
                                "__type": "number",
                                "__value": 0.0,
                                "__range": ((4, 28), (4, 31)),
                            },
                            "__key": {
                                "__type": "string",
                                "__value": "i_am_float_child",
                                "__range": ((4, 8), (4, 26)),
                            },
                        },
                        {
                            "__type": "pair",
                            "__range": ((5, 8), (5, 45)),
                            "__value": {
                                "__type": "string",
                                "__value": "new child value",
                                "__range": ((5, 28), (5, 45)),
                            },
                            "__key": {
                                "__type": "string",
                                "__value": "i_am_float_child",
                                "__range": ((5, 8), (5, 26)),
                            },
                        },
                        {
                            "__type": "pair",
                            "__range": ((6, 8), (9, 9)),
                            "__value": [
                                {
                                    "__type": "pair",
                                    "__range": ((7, 12), (7, 31)),
                                    "__value": {
                                        "__type": "string",
                                        "__value": "child_a",
                                        "__range": ((7, 22), (7, 31)),
                                    },
                                    "__key": {
                                        "__type": "string",
                                        "__value": "_base",
                                        "__range": ((7, 12), (7, 20)),
                                    },
                                },
                                {
                                    "__type": "pair",
                                    "__range": ((8, 12), (8, 22)),
                                    "__value": {
                                        "__type": "number",
                                        "__value": 1,
                                        "__range": ((8, 21), (8, 22)),
                                    },
                                    "__key": {
                                        "__type": "string",
                                        "__value": "hello",
                                        "__range": ((8, 12), (8, 19)),
                                    },
                                },
                            ],
                            "__key": {
                                "__type": "string",
                                "__value": "@child_module_for_test_parser",
                                "__range": ((6, 8), (6, 39)),
                            },
                        },
                    ],
                    "__key": {
                        "__type": "string",
                        "__value": "@child_module_for_test_parser",
                        "__range": ((2, 4), (2, 35)),
                    },
                },
                {
                    "__type": "pair",
                    "__range": ((12, 4), (16, 5)),
                    "__value": [
                        {
                            "__type": "error",
                            "__range": ((13, 8), (15, 21)),
                            "__meta": {
                                "type": "ERROR",
                                "text": '"_base": "child_a",\n        "i_am_float_child": \n        "i_am_child":',
                            },
                        },
                        {
                            "__type": "pair",
                            "__range": ((15, 22), (15, 36)),
                            "__value": None,
                            "__key": {
                                "__type": "string",
                                "__value": "child value2",
                                "__range": ((15, 22), (15, 36)),
                            },
                        },
                    ],
                    "__key": {
                        "__type": "string",
                        "__value": "@child_module_for_test_parser#2",
                        "__range": ((12, 4), (12, 37)),
                    },
                },
                {
                    "__type": "pair",
                    "__range": ((17, 4), (19, 5)),
                    "__value": [
                        {
                            "__type": "pair",
                            "__range": ((18, 8), (18, 36)),
                            "__value": {
                                "__type": "string",
                                "__value": "nest_new_value",
                                "__range": ((18, 20), (18, 36)),
                            },
                            "__key": {
                                "__type": "string",
                                "__value": "nest_key",
                                "__range": ((18, 8), (18, 18)),
                            },
                        }
                    ],
                    "__key": {
                        "__type": "string",
                        "__value": "nested",
                        "__range": ((17, 4), (17, 12)),
                    },
                },
            ],
            "__range": ((0, 0), (20, 1)),
            "__key": {
                "__type": "string",
                "__value": "@module_for_test_parser",
                "__range": ((-1, -1), (-1, -1)),
            },
        }
    ]
    print(root_trace(result, (14, 28)))



# Copyright the author(s) of intc.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
import sys
from os import PathLike, fspath, path
from platform import system
from tempfile import TemporaryDirectory
from typing import List

from setuptools import find_packages, setup

if sys.platform == "win32":
    sys_post_fix = "win"
elif sys.platform == "darwin":
    sys_post_fix = "mac"
else:
    sys_post_fix = "linux"


def write_version_py():
    with open(os.path.join("intc_lsp", "version.txt")) as f:
        version = f.read().strip()

    # write version info to fairseq/version.py
    with open(os.path.join("intc_lsp", "version.py"), "w") as f:
        f.write('__version__ = "{}"\n'.format(version))
    return version


version = write_version_py()


def build_library(output_path: str, repo_paths: List[str]) -> bool:
    """
    Build a dynamic library at the given path, based on the parser
    repositories at the given paths.

    Returns `True` if the dynamic library was compiled and `False` if
    the library already existed and was modified more recently than
    any of the source files.
    """
    output_mtime = path.getmtime(output_path) if path.exists(output_path) else 0

    if not repo_paths:
        raise ValueError("Must provide at least one language folder")

    cpp = False
    source_paths = []
    for repo_path in repo_paths:
        src_path = path.join(repo_path, "src")
        source_paths.append(path.join(src_path, "parser.c"))
        if path.exists(path.join(src_path, "scanner.cc")):
            cpp = True
            source_paths.append(path.join(src_path, "scanner.cc"))
        elif path.exists(path.join(src_path, "scanner.c")):
            source_paths.append(path.join(src_path, "scanner.c"))
    source_mtimes = [path.getmtime(__file__)] + [
        path.getmtime(path_) for path_ in source_paths
    ]

    if max(source_mtimes) <= output_mtime:
        return False

    # local import saves import time in the common case that nothing is compiled
    try:
        from distutils.ccompiler import new_compiler
        from distutils.unixccompiler import UnixCCompiler
    except ImportError as err:
        raise RuntimeError(
            "Failed to import distutils. You may need to install setuptools."
        ) from err

    compiler = new_compiler()
    if isinstance(compiler, UnixCCompiler):
        compiler.set_executables(compiler_cxx="c++")

    with TemporaryDirectory(suffix="tree_sitter_language") as out_dir:
        object_paths = []
        for source_path in source_paths:
            if system() == "Windows":
                flags = None
            else:
                flags = ["-fPIC"]
                if source_path.endswith(".c"):
                    flags.append("-std=c11")
            object_paths.append(
                compiler.compile(
                    [source_path],
                    output_dir=out_dir,
                    include_dirs=[path.dirname(source_path)],
                    extra_preargs=flags,
                )[0]
            )
        compiler.link_shared_object(
            object_paths,
            output_path,
            target_lang="c++" if cpp else "c",
        )
    return True


build_library(
    os.path.join("intc_lsp", "lib", f"json_{sys_post_fix}_ts.so"),
    [os.path.join("intc_lsp", "csrc", "json")],
)
build_library(
    os.path.join("intc_lsp", "lib", f"yaml_{sys_post_fix}_ts.so"),
    [os.path.join("intc_lsp", "csrc", "yaml")],
)

with open("README.md", encoding="utf-8") as f:
    readme = f.read()

with open("LICENSE", encoding="utf-8") as f:
    license = f.read()

setup(
    version=version,
    url="https://github.com/cstsunfu/intc",
)



# Copyright the author(s) of intc.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import time

from tree_sitter import Language, Parser


def get_change(old_source_byte: bytes, new_source_byte: bytes):
    """get the changed range between old source and new source
    Args:
        old_source_byte: the old source
        new_source_byte: the new source
    Returns:
        a dict contain the changed range
    """
    old_byte_lines = old_source_byte.split(b"\n")
    new_byte_lines = new_source_byte.split(b"\n")
    start_line = 0
    start_byte = 0
    old_end_line = len(old_byte_lines)
    new_end_line = len(new_byte_lines)
    old_end_byte = len(old_source_byte)
    new_end_byte = len(new_source_byte)
    for i in range(min(len(old_byte_lines), len(new_byte_lines))):
        if not old_byte_lines[i] == new_byte_lines[i]:
            break
        start_line = i
        start_byte += len(old_byte_lines) + 1

    for i in range(1, min(len(old_byte_lines), len(new_byte_lines))):
        if not old_byte_lines[-i] == new_byte_lines[-i]:
            old_end_line -= i
            new_end_line -= i
            break
        cur_line_byte = len(old_byte_lines[-i]) + 1
        old_end_byte -= cur_line_byte
        new_end_byte -= cur_line_byte
    return {
        "start_byte": start_byte,
        "old_end_byte": old_end_byte,
        "new_end_byte": new_end_byte,
        "start_point": (start_line, 0),
        "old_end_point": (old_end_line, 0),
        "new_end_point": (new_end_line, 0),
    }


if __name__ == "__main__":
    HJSON_LANGUAGE = Language("intc_lsp/lib/json_ts.so", "json")
    parser = Parser()
    parser.set_language(HJSON_LANGUAGE)

    old_source_str = """
    {
        "processor": {
            "_base": "basic@span_cls#pretrained",
            "config": {
                "feed_order": ["load", "seq_lab_loader", "tokenizer", "label_gather", "span_cls_relabel", "save"]
                "tokenizer_config_path": "./data/bert/tokenizer.json", // the tokenizer config path (the tokenizer.json path)
                "data_dir": "./bert/output/",  // save load data base dir
                "size": 3,  // save load data base dir
                "drop": 0.3,  // save load data base dir
            },
        }
    }
    """

    new_source_str = """
    {
        "processor": {
            "_base": "basic@span_cls#pretrained",
                "drop": 0.3,  // save load data base dir
            },
        }
    }
    """

    old_source = old_source_str.encode("utf8")
    new_source = new_source_str.encode("utf8")

    old_tree = parser.parse(old_source)

    ctime = 0
    for _ in range(1000):
        ss = time.time()
        changes = get_change(old_source, new_source)
        old_tree.edit(**changes)
        ctime += time.time() - ss
        new_tree = parser.parse(new_source, old_tree)
    print(f"get diff: {ctime}")

    start = time.time()
    for _ in range(1000):
        changes = get_change(old_source, new_source)
        old_tree.edit(**changes)
        new_tree = parser.parse(new_source, old_tree)
    end = time.time()
    print(f"get diff + add parse: {end - start}")

    start = time.time()
    for _ in range(1000):
        # old_tree.edit(**get_change(old_source_str, new_source_str))
        new_tree = parser.parse(new_source)
    end = time.time()
    print(f"full parse new: {end - start}")

    start = time.time()
    for _ in range(1000):
        # old_tree.edit(**get_change(old_source_str, new_source_str))
        new_tree = parser.parse(old_source)
    end = time.time()
    print(f"full parse old: {end - start}")



# Copyright the author(s) of intc.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
import sys
from typing import Dict, List, Optional, Union

from tree_sitter import Language, Node, Parser

if sys.platform == "win32":
    sys_post_fix = "win"
elif sys.platform == "darwin":
    sys_post_fix = "mac"
else:
    sys_post_fix = "linux"


class JsonParser(object):
    """docstring for JsonParser"""

    def __init__(self):
        super(JsonParser, self).__init__()
        self._parser = Parser()
        dynamic_lib_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
            "lib",
            f"json_{sys_post_fix}_ts.so",
        )
        self._parser.set_language(Language(dynamic_lib_path, "json"))
        # self._parser.set_language(Language('../build/json_ts.so', 'json'))
        self.skip = {"comment", '"', "[", "]", "}", "{", ":"}

    def parser(self, doc: str) -> Optional[Union[Dict, List]]:
        """parser the json string to AST

        Args:
            doc: the source document

        Returns:
            the parser tree
        """
        return self.parser_object(self._parser.parse(bytes(doc, "utf8")).root_node)

    @staticmethod
    def print_parser_tree(node: Node, deep=0) -> None:
        for children in node.named_children:
            print(deep * "    " + children.type)
            JsonParser.print_parser_tree(children, deep + 1)

    def drop_quote(self, _text: bytes) -> str:
        text = _text.decode()
        if (
            len(text) >= 2
            and (text[0] == '"' and text[-1] == '"')
            or (text[0] == "'" and text[-1] == "'")
        ):
            return text[1:-1]
        return text

    def parser_pair(self, node: Node, deep: int) -> Dict:
        key = self.parser_object(node.named_children[0])

        if len(node.named_children) == 2:
            value = self.parser_object(node.named_children[1])
        else:
            value = None
        result = {}
        result["__type"] = "pair"
        result["__range"] = (node.start_point, node.end_point)
        result["__value"] = value
        result["__key"] = key
        return result

    def parser_string(self, node: Node, deep: int) -> Dict:
        return {
            "__type": "string",
            "__value": self.drop_quote(node.text),
            "__range": (node.start_point, node.end_point),
        }

    def parser_number(self, node: Node, deep: int) -> Dict:
        return {
            "__type": "number",
            "__value": eval(node.text.decode()),
            "__range": (node.start_point, node.end_point),
        }

    def parser_array(self, node: Node, deep: int) -> Dict:
        return {
            "__type": "array",
            "__value": [self.parser_object(child) for child in node.named_children],
            "__range": (node.start_point, node.end_point),
        }

    def parser_null_true_false(self, node: Node, deep: int):
        return {
            "__type": node.type,
            "__value": node.text.decode(),
            "__range": (node.start_point, node.end_point),
        }

    def parser_object(self, node, deep=0):
        if node.type in self.skip:
            return None
        if node.type == "pair":
            return self.parser_pair(node, deep + 1)
        if node.type == "string":
            return self.parser_string(node, deep + 1)

        if node.type == "number":
            return self.parser_number(node, deep + 1)

        if node.type == "array":
            return self.parser_array(node, deep + 1)

        if node.type in {"null", "bool"}:
            return self.parser_null_true_false(node, deep + 1)

        if node.type == "object":
            values = []
            for child in node.named_children:
                if child.type in self.skip:
                    continue
                value = self.parser_object(child)
                if value:
                    values.append(value)
            return values
        if node.type == "document":
            result = []
            for child in node.named_children:
                if child.type in self.skip:
                    continue
                result.append(
                    {
                        "__type": "document",
                        "__value": self.parser_object(child, deep + 1),
                        "__range": (child.start_point, child.end_point),
                    }
                )
            return result
        if node.type != "ERROR":
            result = []
            for child in node.named_children:
                if child.type in self.skip:
                    continue
                result.append(
                    {
                        "__type": "document",
                        "__value": self.parser_object(child, deep + 1),
                        "__range": (child.start_point, child.end_point),
                    }
                )
            return result
        return {
            "__type": "error",
            "__range": (node.start_point, node.end_point),
            "__meta": {"type": node.type, "text": node.text.decode()},
        }


if __name__ == "__main__":
    parser = JsonParser()
    print(
        parser.parser(
            """
{
    "_global_":{
        "insert_0": "insert_0",
    },
    "@module_for_test_parser@config_a": {
        "_name_": "config_b",
        "_anchor_": "config_a",
        "@child_module_for_test_parser@child_new": {
            "_name_": "child_module_for_test_parser@child_a",
            "i_am_float_child": 0.0,
            "i_am_child": "new child value"
        },
        "@child_module_for_test_parser@child_b#2": {
            "_name_": "child_module_for_test_parser@child_a",
            "i_am_float_child": 0.0,
            "i_am_child": "child value2"
        }
    }
}
                        """
        )
    )



# Copyright the author(s) of intc.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import pathlib
import re
import urllib
from enum import Enum
from functools import lru_cache
from textwrap import dedent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from urllib.parse import unquote

from attrs import define, field, fields
from lsprotocol.types import (
    CompletionItem,
    CompletionItemKind,
    CompletionItemLabelDetails,
    CompletionList,
    Diagnostic,
    DiagnosticSeverity,
    Hover,
    Location,
    MarkupContent,
    MarkupKind,
    Position,
    Range,
)
from pygls.server import LanguageServer

logger = logging.getLogger("intc_lsp")
try:
    from intc_lsp.src.parser_json import JsonParser
    from intc_lsp.src.parser_yaml import YamlParser
    from intc_lsp.src.trace import root_trace
except Exception as e:
    logger.error(f"import parser error : {e}")

try:
    from intc.register import ic_help, ic_repo, type_module_map
    from intc.utils import split_trace
except Exception as e:
    logger.error(f"import ic_repo error : {e}")


@lru_cache(maxsize=128)
def value2str(value):
    """convert the value to str

    Args:
        value: any type of value
    Returns:
        str value
    """
    if isinstance(value, str):
        return value
    if isinstance(value, bool):
        return str(value).lower()
    if isinstance(value, (list, tuple, dict)):
        return json.dumps(value)
    if value is None:
        return "null"
    try:
        return str(value)
    except:
        return ""


@lru_cache(maxsize=128)
def get_module_type_by_uri(server, uri: str):
    """get the module type by the uri
    Args:
        server: IntcLanguageServer
        uri: the file uri

    Returns:
        module type
        is_entry: True or False, means the uri is a entry module file
    """
    uri = unquote(uri)
    uri_parser_result = urllib.parse.urlparse(uri)
    assert uri_parser_result.scheme == "file"
    uri = uri_parser_result.path
    file_name = pathlib.Path(uri).name
    if server.entry_pattern and server.entry_pattern.match(uri):
        # NOTE: for entry, there is no module_type info from the filename
        return "", True
    elif server.modules_pattern and server.modules_pattern.match(uri):
        module_type = file_name.split(".")[0].split("#")[0].split("@")[0]
        return module_type, False
    return "", False


@define
class ParseResult:
    semantic_trace: str = field(
        default="", metadata={"help": "The semantic trace path, resolve the .name"}
    )
    lex_trace: str = field(default="", metadata={"help": "The cursor node path"})
    trace_result: str = field(default="", metadata={"help": "The cursor node value"})
    is_entry: bool = field(
        default=False,
        metadata={"help": "True means the file is a entry module, False means not"},
    )
    is_key_or_none: Optional[bool] = field(
        default=None,
        metadata={
            "help": "True or False means matching the key or value, None means not match"
        },
    )
    parser_tree: Optional[Dict] = field(
        default=None, metadata={"help": "The parser tree"}
    )
    anchor_dict: Optional[Dict] = field(
        default=None,
        metadata={
            "help": "The anchors dict in the path to the cursor, key is anchor name, value is the trace path"
        },
    )
    additional: Optional[Dict] = field(
        default=None,
        metadata={
            "help": "Additional info, when the cursor is on module root, return the base module name"
        },
    )


class HelpStatus(Enum):
    SUCCESS = "SUCCESS"
    NO_MODULE_NAME = "NO_MODULE_NAME"
    NO_MODULE_HELP = "NO_MODULE_HELP"
    NO_MODULE_TYPE_NAME = "NO_MODULE_TYPE_NAME"
    UN_RECOGNIZED = "UN_RECOGNIZED"


class HoverType(Enum):
    SUCCESS = 0
    VALUE_NOT_IN_OPTION = 1
    MODULE_NOT_FOUND = 2
    CURSOR_WORD_NOT_FOUND = 3
    RESOLVE_ERROR = 4
    UN_COVER_ERROR = 5
    HELP_INFO_NOT_FOUND = 6
    DEPRECATED = 7


def find_module_root(parser_result: ParseResult):
    """find the module name, and the relative trace path

    Args:
        parser_result:
            current cursor node information

    Returns:
        module_name, traces
    """

    def find_module_root_by_parent(
        semantic_traces: List[str], lex_traces: List[str], trace: str
    ):
        """find the real module name based on the parent modules

        Args:
            semantic_traces: the whole semantic trace paths
            lex_traces: the whole lex trace paths
            trace: final trace

        Returns:
            model_name like @module_type@module_name
        """
        par_traces = [trace]
        for semantic_trace, lex_trace in zip(semantic_traces[::-1], lex_traces[::-1]):
            logger.info(
                f"find_module_root_by_parent: semantic_trace : {semantic_trace}, lex_trace : {lex_trace}"
            )
            if (
                semantic_trace.startswith("@")
                and len(semantic_trace.lstrip("@").split("@")) >= 2
            ):
                type_names = semantic_trace.lstrip("@").split("@")
                module_type = type_names[0].split("#")[0]
                module_name = type_names[-1].split("#")[0]
                try:
                    par_module = ic_repo[(module_type, module_name)]
                    for par_trace in par_traces:
                        par_module = par_module[par_trace]
                    base_module_name = par_module.get("_base", "") or par_module.get(
                        "_name", ""
                    )
                    if base_module_name:
                        type_names = trace.lstrip("@").split("@")
                        module_type = type_names[0].split("#")[0]
                        base_module_name = type_names[-1].split("#")[0]
                        return f"@{module_type}@{base_module_name}"
                except Exception as e:
                    logger.info(
                        f"find_module_root_by_parent: can not find the module {module_type}@{module_name}, {e}"
                    )
                    return ""
            if lex_trace.startswith("@"):
                lex_trace = f'@{lex_trace.lstrip("@")}'
            par_traces.insert(0, lex_trace)
        return ""

    try:
        if parser_result.semantic_trace is None:
            return "", []
        semantic_traces = split_trace(parser_result.semantic_trace)
        lex_traces = split_trace(parser_result.lex_trace)
        if len(semantic_traces) != len(lex_traces):
            logger.error(
                f"the length of semantic_traces and lex_traces is not equal, '{parser_result.semantic_trace}', '{parser_result.lex_trace}'"
            )
            lex_traces = semantic_traces
        if (
            parser_result.is_key_or_none == True
            and parser_result.trace_result is not None
        ):
            semantic_traces.append(parser_result.trace_result)
            lex_traces.append(parser_result.trace_result)
        logger.info(f"find_module_root: semantic_traces : {semantic_traces}")
        logger.info(f"find_module_root: lex_traces : {lex_traces}")
        if (
            parser_result.additional
            and "on_semantic_module_name" in parser_result.additional
        ):
            on_semantic_module_name = parser_result.additional[
                "on_semantic_module_name"
            ]
            if (
                on_semantic_module_name.startswith("@")
                and len(on_semantic_module_name.lstrip("@").split("@")) >= 2
            ):
                return parser_result.additional["on_semantic_module_name"], []
        traces = []
        logger.info(f"find_module_root: semantic_traces : {semantic_traces}")
        for i, (trace, lex_trace) in enumerate(
            zip(semantic_traces[::-1], lex_traces[::-1])
        ):
            if trace.startswith("@"):
                if len(trace.lstrip("@").split("@")) >= 2:
                    type_names = trace.lstrip("@").split("@")
                    module_type = type_names[0].split("#")[0]
                    module_name = type_names[-1].split("#")[0]
                    return f"@{module_type}@{module_name}", traces
                else:
                    module_name = find_module_root_by_parent(
                        semantic_traces[::-1][i + 1 :], lex_traces[::-1][i + 1 :], trace
                    )
                    logger.info(
                        f"find_module_root: by parent module_name : {module_name}, traces : {traces}"
                    )
                    if module_name:
                        return module_name, traces
                    else:
                        return trace, traces
            else:
                traces.insert(0, trace)
        return "", []
    except Exception as e:
        logger.error(f"get_key_help: error : {e}")
        return "", []


class IntcResolve(object):
    """docstring for IntcResolve"""

    def __init__(self, server):
        super(IntcResolve, self).__init__()
        self.server: LanguageServer = server
        self.json_parser: JsonParser = None
        self.yaml_parser: YamlParser = None
        self.reserved_words = {"_base", "_name", "_anchor", "_search", "_G"}

        try:
            self.json_parser = JsonParser()
            logger.info(f"init : json parser")
        except Exception as e:
            logger.error(f"init : json parser error : {e}")
        try:
            self.yaml_parser = YamlParser()
            logger.info(f"init : yaml parser")
        except Exception as e:
            logger.error(f"init : yaml parser error : {e}")

    def cursor_word(
        self, uri: str, source: str, position: Position, include_all: bool = True
    ) -> Optional[Tuple[str, Range]]:
        if source is None:
            source = self.server.workspace.get_document(uri).source

        def _cursor_line() -> str:
            line = source.split("\n")[position.line]
            return str(line)

        line = _cursor_line()

        cursor = position.character
        for m in re.finditer(r"\w+", line):
            end = m.end() if include_all else cursor
            if m.start() <= cursor <= m.end():
                word = (
                    line[m.start() : end],
                    Range(
                        start=Position(line=position.line, character=m.start()),
                        end=Position(line=position.line, character=end),
                    ),
                )
                return word
        return None

    def get_module_help(self, module_name: str):
        """get the module help from ic_help

        Args:
            module_name: @module_type@module_name

        Returns:
            help information of the module
        """
        logger.info(f"get_module_help: module_name: {module_name}")
        type_names = module_name.lstrip("@").split("@")
        if len(type_names) < 1:
            logger.error(
                f"get_module_help: unrecognized module_name error : {module_name}"
            )
            return {}, HelpStatus.UN_RECOGNIZED
        elif len(type_names) == 1:
            module_type = type_names[0].split("#")[0]
            name = ""
        else:
            module_type = type_names[0].split("#")[0]
            name = type_names[-1].split("#")[0]
        try:
            return ic_help[(module_type, name)], HelpStatus.SUCCESS
        except Exception as e:
            logger.error(f"get_module_help error: {e}")
            if not module_type and not name:
                return {}, HelpStatus.NO_MODULE_TYPE_NAME
            elif len(type_names) == 1 or not name:
                return {}, HelpStatus.NO_MODULE_NAME
            else:
                return {}, HelpStatus.NO_MODULE_HELP

    def completions(
        self, position: Position, uri: str, trigger_char: str
    ) -> CompletionList:
        """get the completions items for the given position and uri

        Args:
            position: cursor position
            uri: the file uri
            trigger_char: the trigger character

        Returns:
            completions items
        """
        items = []
        source = self.server.workspace.get_document(uri).source

        def _update_source(source):
            lines = source.split("\n")
            line = lines[position.line]
            if not line.strip():
                line = " " * position.character + '" "'
            elif line.strip().endswith(":"):
                rline = line.rstrip()
                lrline = len(rline)
                if position.character > lrline:
                    line = (
                        rline
                        + " " * max((position.character - len(rline) - 1), 0)
                        + '" ",'
                    )
            return "\n".join(
                lines[: position.line] + [line] + lines[position.line + 1 :]
            )

        try:
            source = _update_source(source)
        except Exception as e:
            logger.error(f"completions: update source error : {e}")
            return CompletionList(is_incomplete=False, items=[])
        logger.info(f"completions: source : {source}")
        parser_tree = self.parser_tree(uri, source)

        module_type, is_entry = get_module_type_by_uri(self.server, uri)
        logger.info(f"uri: {uri}. module_type: {module_type}")
        parser_result = self.parser_cursor(
            parser_tree, position, module_type, source, is_entry
        )

        logger.info(f"completions parser_result : {parser_result}")

        # 1. can not resolve the semantic trace
        if parser_result.semantic_trace is None:
            return CompletionList(is_incomplete=False, items=[])
        module_name, traces = find_module_root(parser_result)

        module_help_meta, get_help_status = self.get_module_help(module_name)
        logger.info(
            f"completion: module_help_meta : {module_help_meta}, get_help_status : {get_help_status}, traces : {traces}"
        )
        # 2. can not resolve the module name
        if get_help_status == HelpStatus.NO_MODULE_NAME:
            # 2.2. only module type, no module name, completion on key, complete "_base"
            if parser_result.is_key_or_none == True and not parser_result.trace_result:
                items.append(
                    CompletionItem(
                        label=f"_base",
                        kind=CompletionItemKind.Module,
                        label_details=CompletionItemLabelDetails(
                            description="BaseField"
                        ),
                        documentation=MarkupContent(
                            kind=MarkupKind.Markdown,
                            value=f"BaseField\n\nThe initial module, you can set the base module name as a `str` in its value.",
                        ),
                    )
                )
                return CompletionList(is_incomplete=True, items=items)
            # 2.3. only module type, no module name, completion on value, complete the inherited module name
            elif (parser_result.is_key_or_none == False and traces == ["_base"]) or (
                parser_result.is_key_or_none == True
                and parser_result.trace_result.startswith("@")
                and trigger_char == "@"
            ):
                logger.info(f"completion: the module name is {module_name}")
                module_type = module_name.lstrip("@").split("#")[0].split("@")[0]
                for name in type_module_map[module_type]:
                    module_help = ic_help[(module_type, name)]
                    items.append(
                        CompletionItem(
                            label=name,
                            kind=CompletionItemKind.Module,
                            label_details=CompletionItemLabelDetails(
                                description="ModuleField"
                            ),
                            documentation=MarkupContent(
                                kind=MarkupKind.Markdown,
                                value=module_help.get("_meta", {}).get("help", ""),
                            ),
                        )
                    )
                return CompletionList(is_incomplete=True, items=items)
            # 2.4. trigger by "@" completion on the module type
            elif (
                parser_result.is_key_or_none == True
                and parser_result.trace_result == "@"
            ):
                for module_type in type_module_map:
                    items.append(
                        CompletionItem(
                            label=module_type,
                            label_details=CompletionItemLabelDetails(
                                description="ModuleField"
                            ),
                            kind=CompletionItemKind.Module,
                            documentation=MarkupContent(
                                kind=MarkupKind.Markdown,
                                value=f"ChildModule\n\nThe configure of `{module_type}` module",
                            ),
                        )
                    )
                return CompletionList(is_incomplete=False, items=items)
        # 3. success resolve
        elif get_help_status == HelpStatus.SUCCESS:
            # 3.1. completion on key, complete the module para names
            if parser_result.is_key_or_none == True:
                if not module_help_meta:
                    return CompletionList(is_incomplete=False, items=[])
                for trace in traces:
                    if trace:
                        module_help_meta = module_help_meta.get("properties", {}).get(
                            trace, {}
                        )
                    if not module_help_meta:
                        return CompletionList(is_incomplete=False, items=[])
                properties = module_help_meta.get("properties", {})
                for key in properties:
                    if key in {"_G"} and not parser_result.is_entry:
                        continue
                    help = ""
                    kind = CompletionItemKind.Field
                    if key.startswith("@"):
                        help = "The `Children` module"
                        kind = CompletionItemKind.Class
                        type_name = "SubModule"
                    else:
                        if key in self.reserved_words:
                            kind = CompletionItemKind.Property
                        meta = properties[key]
                        type_name = meta.get("type_name", "")
                        is_deprecated = meta.get("deprecated", False)
                        prefix = ""
                        if is_deprecated == True:
                            prefix = "`Deprecated`\n"
                        elif is_deprecated and isinstance(is_deprecated, str):
                            prefix = f"`Deprecated` {is_deprecated}\n"

                        if type_name:
                            _key = key.replace("_", r"\_")
                            type_str = f"""{_key} -> `{type_name}`\n\n"""
                        else:
                            type_str = ""
                        if key in self.reserved_words:
                            help = f"""{prefix}{type_str}\n{dedent(meta.get("description", ""))}"""
                        else:
                            help = f"""{prefix}{type_str}\n{dedent(meta.get("description", ""))}"""

                    items.append(
                        CompletionItem(
                            label=key,
                            label_details=CompletionItemLabelDetails(
                                description=type_name
                            ),
                            kind=kind,
                            documentation=MarkupContent(
                                kind=MarkupKind.Markdown, value=help
                            ),
                        )
                    )
                return CompletionList(is_incomplete=True, items=items)
            # 3.2. completion the basic module name
            elif parser_result.is_key_or_none == False and traces == ["_base"]:
                module_type = module_name.lstrip("@").split("#")[0].split("@")[0]
                for name in type_module_map[module_type]:
                    module_help_meta = ic_help[(module_type, name)]
                    help = dedent(module_help_meta.get("description", ""))
                    if help:
                        help = f"{module_type}@{name} -> `ModuleField`\r\n{help}"
                    else:
                        help = "ModuleField"
                    items.append(
                        CompletionItem(
                            label=name,
                            kind=CompletionItemKind.Module,
                            label_details=CompletionItemLabelDetails(
                                description="ModuleField"
                            ),
                            documentation=MarkupContent(
                                kind=MarkupKind.Markdown, value=help
                            ),
                        )
                    )
                return CompletionList(is_incomplete=True, items=items)
            # 3.3 completion on value, complete the value suggestions
            elif parser_result.is_key_or_none == False:
                logger.info(f"completion: completion on value, traces : {traces}")
                if not module_help_meta:
                    return CompletionList(is_incomplete=False, items=[])
                for trace in traces:
                    if trace:
                        module_help_meta = module_help_meta.get("properties", {}).get(
                            trace, {}
                        )
                    if not module_help_meta:
                        return CompletionList(is_incomplete=False, items=[])
                meta = module_help_meta
                if not meta:
                    return CompletionList(is_incomplete=False, items=[])
                logger.info(f"completion: completion on value, meta : {meta}")
                default = meta.get("default", "")
                suggestions = meta.get("suggestions", [])
                if default != "":
                    items.append(
                        CompletionItem(
                            label=value2str(default),
                            label_details=CompletionItemLabelDetails(
                                description=meta.get("type", "value")
                            ),
                            kind=CompletionItemKind.Value,
                            documentation=MarkupContent(
                                kind=MarkupKind.Markdown, value=f"Default value"
                            ),
                        )
                    )
                    suggestions = [
                        suggestion
                        for suggestion in suggestions
                        if suggestion != default
                    ]
                options = meta.get("enum", [])
                options = [
                    option for option in options if option not in set(suggestions)
                ]

                for option in options:
                    items.append(
                        CompletionItem(
                            label=value2str(option),
                            kind=CompletionItemKind.Value,
                            label_details=CompletionItemLabelDetails(
                                description=meta.get("type", "value")
                            ),
                            documentation=MarkupContent(
                                kind=MarkupKind.Markdown, value=f"Option Value"
                            ),
                        )
                    )
                for suggestion in suggestions:
                    items.append(
                        CompletionItem(
                            label=value2str(suggestion),
                            kind=CompletionItemKind.Value,
                            label_details=CompletionItemLabelDetails(
                                description=meta.get("type", "value")
                            ),
                            documentation=MarkupContent(
                                kind=MarkupKind.Markdown, value=f"Suggestion Value"
                            ),
                        )
                    )
                logger.info(f"completion: completion on value, items : {items}")
                return CompletionList(is_incomplete=False, items=items)
        elif get_help_status == HelpStatus.NO_MODULE_TYPE_NAME:
            if parser_result.is_key_or_none == True and trigger_char == "@":
                for module_type in type_module_map:
                    items.append(
                        CompletionItem(
                            label=module_type,
                            kind=CompletionItemKind.Module,
                            label_details=CompletionItemLabelDetails(
                                description="ModuleField"
                            ),
                            documentation=MarkupContent(
                                kind=MarkupKind.Markdown,
                                value="ChildModule\n\nThe configure of `{module_type}` module",
                            ),
                        )
                    )
                return CompletionList(is_incomplete=True, items=items)

        return CompletionList(is_incomplete=False, items=items)

    def definition(
        self, position: Position, uri: str, source: str = None
    ) -> List[Location]:
        """get the definition information for the given position and uri

        Args:
            position: cursor position
            uri: the file uri

        Returns:
            list of Location object
        """
        word = self.cursor_word(uri, source, position, True)
        if not word:
            return []

        if source is None:
            source = self.server.workspace.get_document(uri).source
        parser_tree = self.parser_tree(uri, source)
        module_type, is_entry = get_module_type_by_uri(self.server, uri)
        logger.info(f"uri: {uri}. module_type: {module_type}")
        parser_result = self.parser_cursor(
            parser_tree, position, module_type, source, is_entry
        )

        if parser_result.semantic_trace is None:
            return []
        logger.info(f"parser_result : {parser_result}")
        module_name, traces = find_module_root(parser_result)
        logger.info(f"definition: module_name : {module_name}, traces : {traces}")

        if not module_name:
            return []
        module_help_meta, _ = self.get_module_help(module_name)
        if not module_help_meta:
            return []

        logger.info(f"definition: module_help_meta: {module_help_meta}")
        if not traces or traces == ["_base"]:
            file_path = module_help_meta.get("position", {})
            file_paths = []
            if file_path:
                file_paths.append(file_path)
            for inter_file_path in module_help_meta.get("inter_files", []):
                file_paths.append({"file_path": inter_file_path, "line_no": 0})
            return [
                Location(
                    uri=f"file://{file['file_path']}",
                    range=Range(
                        start=Position(line=max(file["line_no"] - 1, 0), character=0),
                        end=Position(line=max(file["line_no"] - 1, 0), character=0),
                    ),
                )
                for file in file_paths
            ]
        for trace in traces:
            module_help_meta = module_help_meta.get("properties", {}).get(trace, {})
        if module_help_meta and "position" in module_help_meta:
            para_position = module_help_meta.get("position", {})
            if not para_position:
                return []
            return [
                Location(
                    uri=f"file://{para_position['file_path']}",
                    range=Range(
                        start=Position(
                            line=max(para_position["line_no"] - 1, 0), character=0
                        ),
                        end=Position(
                            line=max(para_position["line_no"] - 1, 0), character=0
                        ),
                    ),
                )
            ]
        return []

    def hover(self, position: Position, uri: str = "", source: str = None) -> Dict:
        """resolve the hover information for the given position and uri

        Args:
            position: cursor position
            uri: the file uri

        Returns:
            Hover object or None

        """
        word = self.cursor_word(uri, source, position, True)
        logger.info(f"hover: position : {position}, word : {word}")
        if not word:
            return {
                "type": HoverType.CURSOR_WORD_NOT_FOUND,
                "field_type": None,
                "message": "Not found the cursor word",
                "range": Range(
                    start=position,
                    end=position,
                ),
            }
        if source is None:
            source = self.server.workspace.get_document(uri).source
        parser_tree = self.parser_tree(uri, source)
        module_type, is_entry = get_module_type_by_uri(self.server, uri)
        parser_result = self.parser_cursor(
            parser_tree, position, module_type, source, is_entry
        )
        if parser_result.semantic_trace is None:
            logger.error(f"hover parser_result : {parser_result}")
            return {
                "type": HoverType.RESOLVE_ERROR,
                "field_type": None,
                "message": "Can not resolve the semantic trace",
                "range": word[1],
            }
        module_name, traces = find_module_root(parser_result)

        if not module_name:
            return {
                "type": HoverType.RESOLVE_ERROR,
                "field_type": None,
                "message": "Can not resolve the module name",
                "range": word[1],
            }

        if traces and traces[0] in ["_G"]:
            return {
                "type": HoverType.SUCCESS,
                "field_type": None,
                "message": "On the global para",
                "range": word[1],
            }

        module_help_meta, get_help_status = self.get_module_help(module_name)
        if not isinstance(module_help_meta, dict):
            logger.error(
                f"hover: UN_COVER_ERROR, module_help_meta : {module_help_meta}"
            )
            return {
                "type": HoverType.UN_COVER_ERROR,
                "field_type": None,
                "message": "Unrecognized Case",
                "range": word[1],
            }

        if get_help_status != HelpStatus.SUCCESS:
            return {
                "type": HoverType.MODULE_NOT_FOUND,
                "field_type": None,
                "message": "Can not find the module help",
                "range": word[1],
            }

        if (not traces and parser_result.is_key_or_none == True) or (
            traces == ["_base"]
        ):
            return {
                "type": HoverType.SUCCESS,
                "field_type": "ModuleField",
                "message": dedent(
                    f"""
                    {module_name} -> `ModuleField`\r\n
                    {dedent(module_help_meta.get('description', '').strip())}
                    """
                ),
                "range": word[1],
            }
        trace = ""
        for trace in traces:
            module_help_meta = module_help_meta.get("properties", {}).get(trace, {})
        if not isinstance(module_help_meta, dict):
            return {
                "type": HoverType.UN_COVER_ERROR,
                "field_type": None,
                "message": "Unrecognized error",
                "range": word[1],
            }
        if not module_help_meta:
            return {
                "type": HoverType.HELP_INFO_NOT_FOUND,
                "field_type": None,
                "message": "Can not find the help info for this para.",
                "range": word[1],
            }

        def _option_rep(options):
            result = []
            for key in options:
                key = str(key)
                if len(key) > 10:
                    key = key[:8] + "..."
                result.append(f"    {key}")
            return "\n".join(result)

        if trace:
            type_name = module_help_meta.get("type_name", "")
            prefix = ""
            is_deprecated = module_help_meta.get("deprecated", False)
            if is_deprecated == True:
                prefix = "`Deprecated`\n"
            elif is_deprecated and isinstance(is_deprecated, str):
                prefix = f"`Deprecated` {is_deprecated}\n"
            if type_name:
                _trace = trace.replace("_", r"\_")
                type_str = f"""{_trace} -> **`{type_name}`**\n\n"""
            else:
                type_str = ""
            if parser_result.is_key_or_none == True:
                return {
                    "type": HoverType.SUCCESS
                    if not is_deprecated
                    else HoverType.DEPRECATED,
                    "field_type": type_name,
                    "message": f"""{prefix}{type_str}\n{dedent(module_help_meta.get("description", ""))}""",
                    "range": word[1],
                }
            else:
                options_list = module_help_meta.get("enum", [])
                suggest_list = module_help_meta.get("suggestions", [])
                addition_list = module_help_meta.get("additions", [])

                postfix = ""
                if options_list:
                    postfix = f"""`Options`:\n{_option_rep(options_list)}"""
                if suggest_list:
                    postfix = (
                        f"""{postfix}\n`Suggestions`:\n{_option_rep(suggest_list)}"""
                    )
                if addition_list:
                    postfix = (
                        f"""{postfix}\n`Additions`:\n{_option_rep(addition_list)}"""
                    )
                if options_list or addition_list:
                    return {
                        "type": HoverType.SUCCESS,
                        "field_type": type_name,
                        "message": f"""{prefix}{type_str}\n{dedent(module_help_meta.get("description", ""))}\n{postfix}""",
                        "candidate_value": {
                            "options": options_list,
                            "suggestions": suggest_list,
                            "additions": addition_list,
                        },
                        "range": word[1],
                    }
                else:
                    return {
                        "type": HoverType.SUCCESS,
                        "field_type": type_name,
                        "message": f"""{prefix}{type_str}\n{dedent(module_help_meta.get("description", ""))}\n{postfix}""",
                        "range": word[1],
                    }

        logger.error(
            f"hover: UN_COVER_ERROR, module_help_meta : {module_help_meta}, traces: {traces}"
        )
        return {
            "type": HoverType.UN_COVER_ERROR,
            "message": "Unrecognized case",
            "range": word[1],
        }

    def diagnostics(self, uri: str) -> List[Diagnostic]:
        """

        Args:
            uri: the file uri

        Returns:
            diagnostics information

        """
        ignore_keys = {"_anchor", "_G", "_search", "_base"}
        iter_fileds = {"NestField", "SubModule", "ModuleField"}

        @lru_cache(maxsize=64)
        def _(source, uri):
            tree = self.parser_tree(uri, source)

            def _resolve_tree(tree):
                diag_result = []

                if isinstance(tree, list):
                    for sub_tree in tree:
                        diag_result.extend(_resolve_tree(sub_tree))
                elif isinstance(tree, dict):
                    if tree.get("__type") == "document":
                        # detect the base module
                        detected = False
                        try:
                            for sub_tree in tree.get("__value", []):
                                if (
                                    (not isinstance(sub_tree, dict))
                                    or (sub_tree.get("__type", "") != "pair")
                                    or (
                                        sub_tree.get("__key", {}).get("__type")
                                        != "string"
                                    )
                                    or (
                                        sub_tree.get("__key", {}).get("__value")
                                        != "_base"
                                    )
                                ):
                                    logger.warning(
                                        f"diagnostics: skip document value in sub_tree: {sub_tree}"
                                    )
                                    continue
                                base_range = sub_tree.get("__key", {}).get(
                                    "__range", []
                                )
                                base_row, base_column = -1, -1
                                try:
                                    base_row, base_column = (
                                        base_range[0][0],
                                        (base_range[0][1] + base_range[1][1]) // 2,
                                    )
                                except:
                                    logger.warning(
                                        f"diagnostics: can not get base range sub_tree : {sub_tree}"
                                    )
                                    break
                                if base_row == -1 or base_column == -1:
                                    logger.warning(
                                        f"diagnostics: can not get the base_row or base_column, {sub_tree}"
                                    )
                                    break
                                base_position = Position(
                                    line=base_row, character=base_column
                                )
                                hover_result = self.hover(base_position, uri, source)
                                if hover_result["type"] == HoverType.SUCCESS:
                                    detected = True
                                else:
                                    diag_result.append(
                                        Diagnostic(
                                            range=Range(
                                                start=Position(
                                                    line=base_range[0][0],
                                                    character=base_range[0][1],
                                                ),
                                                end=Position(
                                                    line=base_range[1][0],
                                                    character=base_range[1][1],
                                                ),
                                            ),
                                            message="Can not resolve the module name",
                                            severity=DiagnosticSeverity.Error,
                                        )
                                    )
                                break

                        except Exception as e:
                            logger.error(f"diagnostics: detect base error : {e}")
                        try:
                            for sub_tree in tree.get("__value", []):
                                # if can not resolve the module name, ignore the paras for this module
                                if (
                                    not detected
                                    and isinstance(sub_tree, dict)
                                    and (sub_tree.get("__type", "") == "pair")
                                ):
                                    sub_tree_key = sub_tree.get("__key", {})
                                    if not isinstance(sub_tree_key, dict):
                                        logger.error(
                                            f"diagnostics: sub_tree_key : {sub_tree_key}"
                                        )
                                        continue
                                    if not (
                                        sub_tree_key.get("__type", "") == "string"
                                        and sub_tree_key.get("__value", "").startswith(
                                            "@"
                                        )
                                    ):
                                        continue
                                diag_result.extend(_resolve_tree(sub_tree))
                        except Exception as e:
                            logger.error(f"diagnostics: document value error : {e}")
                    elif tree.get("__type", "") == "pair":
                        key = tree.get("__key", {}).get("__value", "")
                        key_range = tree.get("__key", {}).get("__range", [])
                        key_row, key_column = -1, -1
                        # only the key without error, then check the value
                        key_passed = False
                        try:
                            key_row, key_column = (
                                key_range[0][0],
                                (key_range[0][1] + key_range[1][1]) // 2,
                            )
                        except:
                            return diag_result
                        if key_row == -1 or key_column == -1:
                            return diag_result
                        if key.startswith("@"):
                            key_position = Position(line=key_row, character=key_column)
                            hover_result = self.hover(key_position, uri, source)
                            if hover_result["type"] == HoverType.SUCCESS:
                                key_passed = True
                            else:
                                diag_result.append(
                                    Diagnostic(
                                        range=Range(
                                            start=Position(
                                                line=key_range[0][0],
                                                character=key_range[0][1],
                                            ),
                                            end=Position(
                                                line=key_range[1][0],
                                                character=key_range[1][1],
                                            ),
                                        ),
                                        message=hover_result["message"],
                                        severity=DiagnosticSeverity.Error,
                                    )
                                )
                        elif key in ignore_keys:
                            return diag_result
                        else:
                            key_position = Position(line=key_row, character=key_column)
                            hover_result = self.hover(key_position, uri, source)
                            if hover_result["type"] == HoverType.SUCCESS:
                                key_passed = True
                            elif hover_result["type"] == HoverType.DEPRECATED:
                                diag_result.append(
                                    Diagnostic(
                                        range=Range(
                                            start=Position(
                                                line=key_range[0][0],
                                                character=key_range[0][1],
                                            ),
                                            end=Position(
                                                line=key_range[1][0],
                                                character=key_range[1][1],
                                            ),
                                        ),
                                        message="This para is `deprecated`",
                                        severity=DiagnosticSeverity.Information,
                                    )
                                )
                            else:
                                logger.error(
                                    f"diagnostics: hover_result : {hover_result}"
                                )
                                diag_result.append(
                                    Diagnostic(
                                        range=Range(
                                            start=Position(
                                                line=key_range[0][0],
                                                character=key_range[0][1],
                                            ),
                                            end=Position(
                                                line=key_range[1][0],
                                                character=key_range[1][1],
                                            ),
                                        ),
                                        message=hover_result["message"],
                                        severity=DiagnosticSeverity.Error,
                                    )
                                )
                        if key_passed and hover_result["field_type"] in iter_fileds:
                            if not isinstance(tree, dict):
                                return diag_result
                            for sub_tree in tree.get("__value", []):
                                # only check the pairs
                                if (
                                    (not isinstance(sub_tree, dict))
                                    or (not (sub_tree.get("__type", "") == "pair"))
                                    or (
                                        not (
                                            sub_tree.get("__key", {}).get("__type")
                                            == "string"
                                        )
                                    )
                                ):
                                    continue
                                diag_result.extend(_resolve_tree(sub_tree))
                        else:
                            if not isinstance(tree, dict):
                                logger.info(f"diagnostics: not dict tree : {tree}")
                                return diag_result
                            # FIXME: options for list or dict value
                            try:
                                value = tree.get("__value", {}).get("__value", "")
                            except:
                                logger.warning(
                                    f"diagnostics: currently do not support list or dict value"
                                )
                                return diag_result
                            value_range = tree.get("__value", {}).get("__range", [])
                            value_row, value_column = -1, -1
                            try:
                                value_row, value_column = (
                                    value_range[0][0],
                                    value_range[0][1],
                                )
                            except:
                                return diag_result
                            value_position = Position(
                                line=value_row, character=value_column + 1
                            )
                            hover_result = self.hover(value_position, uri, source)
                            options = hover_result.get("candidate_value", {}).get(
                                "options", []
                            )
                            if options:
                                options = set(options)
                                if True in options:
                                    options.remove(True)
                                    options.add("true")
                                if False in options:
                                    options.remove(False)
                                    options.add("false")
                                if None in options:
                                    options.remove(None)
                                    options.add("null")
                                if value not in options:
                                    diag_result.append(
                                        Diagnostic(
                                            range=Range(
                                                start=Position(
                                                    line=value_range[0][0],
                                                    character=value_range[0][1],
                                                ),
                                                end=Position(
                                                    line=value_range[1][0],
                                                    character=value_range[1][1],
                                                ),
                                            ),
                                            message=f"{value} is not in supported options: `{options}`",
                                            severity=DiagnosticSeverity.Warning,
                                        )
                                    )
                    else:
                        logger.warning(f"diagnostics: not document or pair: {tree}")
                else:
                    logger.warning(f"diagnostics: not dict or list tree : {tree}")

                return diag_result

            return _resolve_tree(tree)

        source = self.server.workspace.get_document(uri).source
        return _(source, uri)

    def parser_tree(self, uri: str, source: str = ""):
        """parser the source to AST

        Args:
            uri: the source code uri

        Kwargs:
            source: optional, the source code

        Returns:
            the parser tree
        """

        @lru_cache(maxsize=128)
        def _(uri, source):
            if source is None:
                source = self.server.workspace.get_document(uri).source
            if (
                uri.endswith(".json")
                or uri.endswith(".jsonc")
                or uri.endswith(".json5")
                or uri.endswith(".hjson")
            ):
                try:
                    parser_tree = self.json_parser.parser(source)
                    return parser_tree
                except Exception as e:
                    logger.error(f"parser tree: json parser error : {e}")
                    return {}
            elif uri.endswith(".yaml") or uri.endswith(".yml"):
                try:
                    parser_tree = self.yaml_parser.parser(source)
                    return parser_tree
                except Exception as e:
                    logger.error(f"parser tree: yaml parser error : {e}")
                    return {}
            else:
                logger.warning(f"not support file type {uri}")
                return {}

        return _(uri, source)

    def parser_cursor(
        self,
        parser_tree,
        position: Position,
        module_type_from_uri: str,
        source: str,
        is_entry: bool,
    ) -> ParseResult:
        """parser the source to AST, and get the cursor position node information, and the trace information
        Args:
            parser_tree: the tree of parser result
            position: (line, character)
            module_type_from_uri: the module type detected from the uri
            is_entry: True means the file is a entry module, False means is a submodule
        Returns:
            the ParseResult object
        """
        _position: tuple = (position.line, position.character)

        @lru_cache(maxsize=128)
        def wrap(source, _position, module_type_from_uri, is_entry):
            result = {"is_entry": is_entry}
            result["parser_tree"] = parser_tree
            try:
                (
                    semantic_trace,
                    lex_trace,
                    trace_result,
                    is_key_or_none,
                    anchor_dict,
                    additional,
                ) = root_trace(parser_tree, _position, module_type_from_uri)
            except Exception as e:
                logger.error(f"hover: error : {e}")
                return ParseResult(**result)
            result["semantic_trace"] = semantic_trace
            result["lex_trace"] = lex_trace
            result["trace_result"] = trace_result
            result["is_key_or_none"] = is_key_or_none
            result["anchor_dict"] = anchor_dict
            result["additional"] = additional
            return ParseResult(**result)

        return wrap(source, _position, module_type_from_uri, is_entry)


if __name__ == "__main__":
    resolver = IntcResolve(LanguageServer("test", "0.1"))
